unit Dicty;
(***************************************************************
                  SIMPLE DICTIONARY CLASS
****************************************************************
The dictionary stores variants against a keyword, so for example
D.setInteger('Height',45) stores an integer under the tag 'Height'
and TF := D.GetBoolean('DoneThisBefore?') get a previously
stored boolean.

To create a dictionary:       D := TDictionary.new;
To add items                  D.Set...(<key>,<value>);
To get items                  result := D.Get...(<key>);
To see if changed             if D.Changed ...

The complete dictionary can be saved to and read from file:
  D.SaveToFile(<filename>);    (Erases previous file if any)
  D.LoadFromFile(<filename>);  (Erases previous dictionary entries)

The types supported in get and set are:
     Boolean,Integer,String,Single,Date
Don't forget that Tmemo.text is a string

Trying to save an item as one type and retrieve it as another
will raise an EDictionary error.

Trying to retrieve an item for a key that isn't in the dictionary
will raise an EDictionary error.  So you might have calling
code that looks like this:
  try
    S := D.GetString('Title');
  except
    S := '';
  end;

Keys are case sensitive.

Don't forget to call the free procedure after use.

*)


interface

uses Lists,Classes,Sysutils,useful,dataio;

type

  EDictionary = class(Exception);

  TDictionaryElement = Class(Tobject)
  public
    ky : String;
    val : Variant;
    constructor New (const key:string);
  end;

  TDictionary = Class(TObject)
  private
    L : TList;
    Function GetandCheckVariant(const Ky:String;vtype:integer):Variant;
    function Locate(const Srchky:string):integer;
  public
    Changed : boolean;
    constructor New;
    destructor Destroy ; override;
    Function NoOfEntries:Integer;
    Function KeyAt(const index:integer):string;
    Function VarAt(const index:integer):Variant;
    procedure DeleteItem(const Ky:string);
    Procedure Empty;
    Procedure LoadFromFile(const AFileName:String);
    procedure SaveToFile(const AFileName:String);
    Procedure SetInteger(const Ky:String;const I:Integer);
    Procedure SetString(const Ky:String;const S:String);
    Procedure SetSingle(const Ky:String;const S:Single);
    Procedure SetDate(const Ky:String;const D:TDateTime);
    Procedure SetBoolean(const Ky:String;const TF:Boolean);
    procedure SetVariant(const ky:string;const v:Variant;const vtype:integer);
    Function GetBoolean(const Ky:String):Boolean;
    Function GetInteger(const Ky:String):Integer;
    Function GetString(const Ky:String):String;
    Function GetSingle(const Ky:String):Single;
    Function GetDate(const Ky:String):TDateTime;
    Function GetVariant(const Ky:String):Variant;
  end;


implementation

Function TDictionary.NoOfEntries:Integer;
begin
  result := L.count;
end;

Constructor TDictionaryElement.new(const key:string);
begin
  inherited Create;
  ky := key;
  Val := Unassigned;
end;

Constructor TDictionary.new;
begin
  inherited Create;
  L := TList.create;
  Changed := false;
end;

Destructor TDictionary.Destroy;
begin
  destroylistanditems(L);
  inherited destroy;
end;



function TDictionary.Locate(const Srchky:string):integer;
var
  i : integer;
begin
  result := -1;
  if L.count > 0 then begin
    for i := 0 to L.count - 1 do begin
      if L[i] <> nil then begin
        if TDictionaryElement(L[i]).ky = Srchky then begin
          result := i;
          break;
        end;
      end;
    end;
  end;
  if result = - 1 then begin
    raise EDictionary.createfmt('Key "%s" not found.',[Srchky]);
  end;

end;


Function TDictionary.GetVariant(const Ky:String):Variant;
var
  i : integer;
begin
  i := locate(ky);  // will throw up a can't find error
  Result := TDictionaryElement(L[i]).Val;
end;


Function TDictionary.GetandCheckVariant(const Ky:String;vtype:integer):Variant;
var
  v : variant;
begin
  v := getvariant(ky);
  if vartype(v) <> vtype then begin
    raise EDictionary.createfmt('Dictionary item "%s" is not expected type',[ky]);
  end;
  result := v;
end;


Function TDictionary.GetBoolean(const Ky:String):Boolean;
begin
  result := varastype(getandcheckVariant(ky,VarBoolean),varBoolean);
end;

Function TDictionary.GetString(const Ky:String):String;
begin
  result := varastype(getandcheckVariant(ky,VarString),varString);
end;

Function TDictionary.GetSingle(const Ky:String):Single;
begin
  result := varastype(getandcheckVariant(ky,VarSingle),varSingle);
end;

Function TDictionary.GetInteger(const Ky:String):Integer;
begin
  result := varastype(getandcheckVariant(ky,VarInteger),varInteger);
end;

Function TDictionary.GetDate(const Ky:String):TDateTime;
begin
  result := varastype(getandcheckVariant(ky,VarDate),varDate);
end;




procedure TDictionary.SetVariant(const ky:string;const v:Variant;const vtype:integer);
var
  i : integer;
  E : TDictionaryElement;
begin

  try  // got an entry for this key already?
    i := locate(ky);
  except       // throws an error if not already there
    i := -1;
  end;
  if i < 0 then begin
    E := TDictionaryElement.new(ky);
    i := L.add(E);
  end;

  TDictionaryElement(L[i]).Val := varastype(v,vtype);

  changed := true;

end;

Procedure TDictionary.SetInteger(const Ky:String;const I:Integer);
var
  V:Variant;
begin
  V := i;
  SetVariant(ky,v,varinteger)
end;

Procedure TDictionary.SetString(const Ky:String;const S:String);
var
  V:Variant;
begin
  V := S;
  SetVariant(ky,v,varstring)
end;

Procedure TDictionary.SetSingle(const Ky:String;const S:Single);
var
  V:Variant;
begin
  V := S;
  SetVariant(ky,v,varsingle)
end;

Procedure TDictionary.SetDate(const Ky:String;const D:TDateTime);
var
  V:Variant;
begin
  V := D;
  SetVariant(ky,v,vardate)
end;

Procedure TDictionary.SetBoolean(const Ky:String;const TF:Boolean);
var
  V:Variant;
begin
  V := TF;
  SetVariant(ky,v,varboolean)
end;

procedure TDictionary.SaveToFile(const AFileName:String);
var
  i : integer;
  E : TDictionaryElement;
  NoRecs : integer;
begin
  NoRecs := L.count;
  with TKeyDataFile.Create do begin
    TRY
      OpenForWrite(aFileName,NoRecs);
      for i := 1 to NoRecs do begin
        if l[i]<>nil then begin
          E := L[i];
          WriteNext(E.ky,E.val);
        end;  
      end;
    FINALLY
      free;
    END;
  end;
end;



Procedure TDictionary.LoadFromFile(const AFileName:String);
var
  notoread,i : integer;
  F : textfile;
  vt : integer;
  V : Variant;
  ky,S ,VS: string;
begin
  destroylistitems(L);           // fresh start
  if not fileexists(AFileName) then exit;

  with TKeyDataFile.Create do begin
    TRY
      notoread := OpenForRead(aFileName);
      for i := 1 to notoread do begin
        if ReadNext(ky,v) then
         if ky <> '*NIL*' then SetVariant(ky,v,vartype(v))
        else break;
      end;
    FINALLY
      free;
    END;
  end;

end;


Procedure TDictionary.Empty;
begin
  destroylistitems(L);
end;

procedure TDictionary.DeleteItem(const Ky:string);
var
  i : integer;
begin
  try
    i := Locate(Ky);
    if L.Items[i] <> nil then begin
      TDictionaryElement(L.Items[i]).Free;
      L.delete(i);
      //L.pack;
    end;
  except;  // ignore locate not found error
  end;
end;


Function TDictionary.KeyAt(const index:integer):string;
{ return the key at the specified index ('' if a problem)}
begin
  if (index < 0) or (index > NoOfEntries) then begin
    result := '';
  end else begin
    result := TDictionaryElement(L.Items[index]).ky;
  end;
end;

Function TDictionary.VarAt(const index:integer):Variant;
{ return the variant at the specified index
  return Null if a problem}
begin
  if (index < 0) or (index > NoOfEntries) then begin
    result := Null;
  end else begin
    result := TDictionaryElement(L.Items[index]).val;
  end;
end;


end.
